# -*- coding: utf-8 -*-
import os.path
import sys
from typing import Dict, List

from ApplicationConfig import ApplicationConfig
from Composer import createPrintComposer, upsertComposerTemplatesInLayoutManager
from GisDataState import SITE_CLUSTER_DEFAULT, GisDataState
from Helper import containsFileWithExtension
from Layer import (
    DEFAULT_LAYER_STYLE_NAME,
    changeLayerStyles,
    isNonRevisionRimoLayer,
    refreshLayers,
)
from MessageAggregator import MessageAggregator
from Project import (
    hasRimoData,
    isAProjectCurrentlyOpened,
    loadRimoProject,
    saveCurrentProject,
)
from ProjectHelper import (
    createProjectFilePath,
    createProjectName,
    getExistingProjectFilePath,
)
from ProjectState import ProjectState
from ProjectStateRepository import ProjectStateRepository
from ProjectStateValidator import ProjectStateValidator
from qgis.core import QgsMapLayerType, QgsProject
from qgis.PyQt.QtCore import QCoreApplication
from qgis.PyQt.QtGui import QIcon
from qgis.PyQt.QtWidgets import QAction, QFileDialog
from qgis.utils import reloadPlugin
from Result import Result
from RimoClient import (
    getClusterData,
    getRegions,
    getRimoComposerData,
    getSiteClusterData,
    sendLoginRequest,
)
from RimoProject import createRimoProject
from SessionState import SessionState
from UserMessage import renderInfo, renderWarning

sys.path.append("./requests")
import json
import logging

import requests

# Initialize Qt resources from file resources.py
from .resources import *

# Import the code for the dialog
from .RimoPrinting_dialog import *


class RimoPrinting:
    def __init__(self, qgisInterface):
        self.qgisInterface = qgisInterface

        self.currentQgisProject = QgsProject.instance()

        self.currentQgisProject.readProject.connect(self.onProjectRead)

        self.appConfig = None
        self.initState()

        self.dlg = RimoPrintingDialog()

        self.actions = []
        self.menu = self.tr("&RimoPrinting")

    def initGui(self):
        """Create the menu entries and toolbar icons inside the QGIS GUI.
        Is executed during startup of QGIS"""

        if self.appConfig is not None:
            parent = os.path.basename(self.appConfig.pluginDirPath)

            self.add_action(
                self.appConfig.icon_path,
                text=self.tr(parent),
                callback=self.showRimoDialog,
            )

            self.initUI()
            self.setUIHooks()

    def initUI(self):
        self.dlg.env_comboBox.setCurrentIndex(0)
        self.dlg.username_lineEdit.clear()
        self.dlg.password_lineEdit.clear()
        self.dlg.loginInfo_label.clear()
        self.dlg.tenant_comboBox.clear()
        self.dlg.region_comboBox.clear()
        self.dlg.cluster_comboBox.clear()
        self.dlg.siteCluster_comboBox.clear()

        self.dlg.layer_comboBox.clear()
        self.fillLayerSettingComboBox()

        self.dlg.currentStyle_lineEdit.setText(DEFAULT_LAYER_STYLE_NAME)
        self.dlg.printingType_comboBox.setCurrentIndex(0)
        self.dlg.printingFormat_comboBox.setCurrentIndex(0)
        self.dlg.printingScale_comboBox.setCurrentIndex(0)

        self.setEnablementOfTenantUIElements(False)
        self.setEnablementOfAllImportUIElements(False)
        self.setEnablementOfPrintingElements(False)
        self.dlg.reload_pushButton.setEnabled(False)

    def initState(self):
        self.sessionState = SessionState()
        self.gisDataState = GisDataState()
        self.projectState = ProjectState()

        infoMessageAggregator = MessageAggregator()
        appConfigResult = ApplicationConfig.create(infoMessageAggregator)
        if appConfigResult.isFailure():
            renderWarning(
                f"Rimo Printing plugin can not be loaded due to:\n {appConfigResult.message}",
                self.qgisInterface,
            )
            return
        if infoMessageAggregator.hasMessages():
            renderInfo(
                "\n".join(infoMessageAggregator.getMessages()).strip(),
                self.qgisInterface,
            )
        self.appConfig = appConfigResult.value

        logging.basicConfig(
            format="%(levelname)s[%(lineno)d]  %(filename)s  %(funcName)s():\t%(message)s",
            level=logging.ERROR if self.appConfig.isRunningInProd() else logging.DEBUG,
        )

        if not os.path.exists(self.appConfig.pluginStyleFolderPath):
            os.makedirs(self.appConfig.pluginStyleFolderPath)

    # Setup which has to be done on a project not in general for the plugin
    def setupForIndividualProject(self):
        upsertComposerTemplatesInLayoutManager(
            self.appConfig.getAdditionalComposerTemplatePaths(), self.currentQgisProject
        )

    def setUIHooks(self):
        self.dlg.login_pushButton.clicked.connect(
            lambda: self.login(
                self.dlg.env_comboBox.currentText(),
                self.dlg.username_lineEdit.text(),
                self.dlg.password_lineEdit.text(),
            )
        )

        self.dlg.selectTenant_pushButton.clicked.connect(
            lambda: self.setTenant(
                self.sessionState.tenants[self.dlg.tenant_comboBox.currentIndex()][
                    "tenant"
                ]["asOop"],
                self.sessionState.tenants[self.dlg.tenant_comboBox.currentIndex()][
                    "tenant"
                ]["id"],
                self.sessionState.token,
                self.dlg.env_comboBox.currentText(),
            )
        )

        self.dlg.region_comboBox.activated.connect(
            lambda: self.regionComboBoxChange(
                self.dlg.env_comboBox.currentText(),
            )
        )
        self.dlg.cluster_comboBox.activated.connect(
            lambda: self.clusterComboBoxChange(
                self.dlg.env_comboBox.currentText(),
            )
        )

        self.dlg.layer_comboBox.activated.connect(
            lambda: self.updateLayerSetting(self.dlg.layer_comboBox.currentText())
        )

        self.dlg.load_pushButton.clicked.connect(
            lambda: self.load(
                self.dlg.region_comboBox.currentText(),
                self.dlg.cluster_comboBox.currentText(),
                self.dlg.siteCluster_comboBox.currentText(),
                self.dlg.env_comboBox.currentText(),
                self.dlg.layer_comboBox.currentText(),
                self.dlg.tenant_comboBox.currentText(),
            )
        )

        self.dlg.reload_pushButton.clicked.connect(lambda: self.reloadDataFromRimo())

        self.dlg.selectStyle_pushButton.clicked.connect(
            lambda: self.handleSelectStyle()
        )

        self.dlg.exportQMLs_pushButton.clicked.connect(lambda: self.handleExportQMLs())

        self.dlg.createPrintComposer_pushButton.clicked.connect(
            lambda: self.handleCreatePrintComposer()
        )

    def onProjectRead(self):
        if hasRimoData(self.currentQgisProject):
            projectState = ProjectStateRepository.loadProjectState(
                self.currentQgisProject
            )
            infoMessageAggregator = MessageAggregator()
            projectValidResult = ProjectStateValidator.validate(
                projectState, self.appConfig, infoMessageAggregator
            )
            if infoMessageAggregator.hasMessages():
                renderInfo(
                    "\n".join(infoMessageAggregator.getMessages()).strip(),
                    self.qgisInterface,
                )
            if projectValidResult.isFailure():
                renderWarning(projectValidResult.message, self.qgisInterface)
                return

            self.projectState = projectState

            self.setupForIndividualProject()

            self.dlg.env_comboBox.setCurrentText(self.projectState.environment)
            self.dlg.tenant_comboBox.addItem(self.projectState.tenant)
            self.dlg.region_comboBox.addItem(self.projectState.regionName)
            self.dlg.cluster_comboBox.addItem(self.projectState.clusterName)
            self.dlg.siteCluster_comboBox.addItem(self.projectState.siteClusterName)
            self.dlg.layer_comboBox.addItem(self.projectState.layerSettingName)
            self.dlg.currentStyle_lineEdit.setText(self.projectState.currentStyling)
            self.dlg.env_comboBox.setEnabled(False)
            self.setEnablementOfTenantUIElements(False)
            self.setEnablementOfAllImportUIElements(False)
            self.dlg.layer_comboBox.setEnabled(False)
            self.setEnablementOfPrintingElements(self.sessionState.isLoggedIn())
            self.dlg.reload_pushButton.setEnabled(self.sessionState.isLoggedIn())

    def showRimoDialog(self):
        if self.dlg.isHidden():
            self.dlg.show()

        if not self.dlg.isActiveWindow():
            self.dlg.activateWindow()

        # unfortunately there is no signal for "newProjectCreated or projectCreated like we use for projectOpened" therefore we do it here instead
        if not hasRimoData(self.currentQgisProject):
            if self.sessionState.isLoggedIn():
                self.dlg.layer_comboBox.setEnabled(True)
                self.setEnablementOfTenantUIElements(True)
                self.setEnablementOfAllImportUIElements(True)
                self.setEnablementOfPrintingElements(False)
                self.dlg.reload_pushButton.setEnabled(False)
            else:
                self.resetPlugin()

    def load(
        self,
        regionName: str,
        clusterName: str,
        siteClusterName: str,
        environment: str,
        layerSettingName: str,
        tenant: str,
    ):
        if isAProjectCurrentlyOpened(self.currentQgisProject):
            renderWarning(
                "There is already data or at least one print composer in this project. Please use an empty QGIS-Project to load rimo data!",
                self.qgisInterface,
            )
            return

        projectName = createProjectName(
            clusterName, siteClusterName, environment == "Development"
        )
        qgisProjectExistsResult = getExistingProjectFilePath(
            self.appConfig.getProjectDirPath(projectName), projectName
        )

        if qgisProjectExistsResult.isSuccess():
            renderInfo("Loading existing QGIS project...", self.qgisInterface)
            infoMessageAggregator = MessageAggregator()

            try:
                projectFilePathOfProjectToBeLoaded = qgisProjectExistsResult.value

                loadProjectResult = loadRimoProject(
                    projectFilePathOfProjectToBeLoaded,
                    self.currentQgisProject,
                    self.appConfig,
                    self.qgisInterface,
                    infoMessageAggregator,
                )
                if loadProjectResult.isFailure():
                    renderWarning(loadProjectResult.message, self.qgisInterface)
                    return
                if infoMessageAggregator.hasMessages():
                    renderInfo(
                        "\n".join(infoMessageAggregator.getMessages()).strip(),
                        self.qgisInterface,
                    )

                self.projectState = loadProjectResult.value

            except Exception as e:
                logging.error(e)
                renderWarning(
                    f"Loading of existing QGIS project failed with error:\n{self.currentQgisProject.error()}...",
                    self.qgisInterface,
                )
                return

            refreshLayers(
                self.currentQgisProject,
                self.appConfig,
                self.sessionState.token,
                self.qgisInterface,
                self.projectState,
            )

            infoMessageAggregator = MessageAggregator()
            saveResult = saveCurrentProject(
                self.currentQgisProject,
                projectFilePathOfProjectToBeLoaded,
                self.projectState,
                self.appConfig,
                infoMessageAggregator,
            )
            if infoMessageAggregator.hasMessages():
                renderInfo(
                    "\n".join(infoMessageAggregator.getMessages()).strip(),
                    self.qgisInterface,
                )
            if saveResult.isFailure():
                renderWarning(saveResult.message, self.qgisInterface)
                return

        else:
            self.projectState = ProjectState(
                environment,
                regionName,
                clusterName,
                siteClusterName,
                layerSettingName,
                self.appConfig.getDefaultStylingForLayerSettingName(layerSettingName),
                tenant,
            )
            infoMessageAggregator = MessageAggregator()
            createResult = createRimoProject(
                self.currentQgisProject,
                self.projectState,
                self.appConfig,
                self.sessionState,
                self.gisDataState,
                self.qgisInterface,
                infoMessageAggregator,
            )

            if infoMessageAggregator.hasMessages():
                renderInfo(
                    "\n".join(infoMessageAggregator.getMessages()).strip(),
                    self.qgisInterface,
                )
            if createResult.isFailure():
                renderWarning(createResult.message, self.qgisInterface)
                return

            self.setupForIndividualProject()

        self.dlg.currentStyle_lineEdit.setText(self.projectState.currentStyling)
        self.setEnablementOfPrintingElements(True)
        self.setEnablementOfAllImportUIElements(False)
        self.setEnablementOfTenantUIElements(False)
        self.dlg.reload_pushButton.setEnabled(True)
        self.dlg.layer_comboBox.setEnabled(False)

        self.dlg.close()

    def login(self, environment, userName, password):
        try:
            data = sendLoginRequest(
                self.appConfig.getRimoUrl(environment),
                userName,
                password,
                self.appConfig.pluginVersion,
            )
            if "_error" in data:
                renderWarning(data["_error"].replace("Error:", ""), self.qgisInterface)
                return

            self.sessionState.token = data["token"]
            logging.debug(f"Current Token: {self.sessionState.token}")
        except Exception as e:
            renderWarning("Something went wrong", self.qgisInterface)
            logging.error(e)
            return

        # this replaces the token in all layers with the new token
        if hasRimoData(self.currentQgisProject):
            refreshLayers(
                self.currentQgisProject,
                self.appConfig,
                self.sessionState.token,
                self.qgisInterface,
                self.projectState,
            )
            self.setEnablementOfPrintingElements(True)
            self.dlg.reload_pushButton.setEnabled(True)
        else:
            self.sessionState.tenants = data["userProfile"]
            self.dlg.selectTenant_pushButton.setEnabled(True)
            self.dlg.tenant_comboBox.setEnabled(True)

            tenantList = [
                tenant["tenant"]["id"] for tenant in self.sessionState.tenants
            ]
            self.dlg.tenant_comboBox.clear()
            self.dlg.tenant_comboBox.addItems(tenantList)
            if len(tenantList) == 1:
                self.setTenant(
                    self.sessionState.tenants[0]["tenant"]["asOop"],
                    self.sessionState.tenants[0]["tenant"]["id"],
                    self.sessionState.token,
                    environment,
                )

        self.sessionState.loginStatus = True
        self.dlg.loginInfo_label.setText("Logged In.")

    def setTenant(self, tenantOop: str, tenantId: str, token: str, environment):
        self.sessionState.tenantOop = tenantOop
        self.sessionState.tenantId = tenantId
        self.sessionState.token = token
        self.sessionState.userOop = [
            x
            for x in self.sessionState.tenants
            if x["tenant"]["asOop"] == self.sessionState.tenantOop
        ][0]["asOop"]

        self.dlg.login_pushButton.setEnabled(False)
        self.dlg.region_comboBox.clear()
        self.dlg.cluster_comboBox.clear()
        self.dlg.siteCluster_comboBox.clear()

        regionsResult = self.loadRegions(environment)
        if regionsResult.isFailure():
            renderWarning(
                f"Loading region data for tenant {repr(self.sessionState.tenantId)} failed",
                self.qgisInterface,
            )
            return
        elif not regionsResult.value:
            renderWarning(
                f"No regions for tenant {self.sessionState.tenantId} found",
                self.qgisInterface,
            )
            return

        self.gisDataState.setRegionData(regionsResult.value)
        self.dlg.region_comboBox.addItems(self.gisDataState.getRegionNames())

        clustersResult = self.loadClusters(environment)
        if clustersResult.isFailure():
            renderWarning(
                f"Failed to load cluster data for region {self.dlg.region_comboBox.currentText()}",
                self.qgisInterface,
            )
            return
        elif not clustersResult.value:
            renderWarning(
                f"No cluster data found for region {self.dlg.region_comboBox.currentText()}",
                self.qgisInterface,
            )
            return

        self.gisDataState.setClusterData(clustersResult.value)
        self.dlg.cluster_comboBox.addItems(self.gisDataState.getClusterNames())

        siteClustersResult = self.loadSiteClusters(environment)
        if siteClustersResult.isSuccess():
            self.gisDataState.setSiteClusterData(siteClustersResult.value)
            siteClusterNames = self.gisDataState.getSiteClusterNames()
            siteClusterNames.sort(key=str.lower)
            siteClusterNames.insert(0, SITE_CLUSTER_DEFAULT)
            self.dlg.siteCluster_comboBox.addItems(siteClusterNames)
        else:
            self.dlg.siteCluster_comboBox.addItems([SITE_CLUSTER_DEFAULT])

        self.setEnablementOfAllImportUIElements(True)

    def updateLayerSetting(self, layerSettingName: str):
        result = self.appConfig.canLayerSettingBeSelected(layerSettingName)

        if result.isFailure():
            renderWarning(
                f"{result.message} There is a problem with the configuration file. Consider contacting an administrator.",
                self.qgisInterface,
            )
        else:
            layerSetting = self.appConfig.getLayerSettingWithName(layerSettingName)
            assert layerSetting is not None, "layer setting can not be null"

            self.projectState.layerSettingName = layerSettingName
            self.projectState.currentStyling = layerSetting.style

            self.dlg.layer_comboBox.setCurrentText(layerSettingName)
            self.dlg.currentStyle_lineEdit.setText(self.projectState.currentStyling)

    def handleCreatePrintComposer(self):
        frameData = getRimoComposerData(
            self.appConfig.getRimoUrl(self.projectState.environment),
            self.sessionState.tenantId,
            self.gisDataState.getOopForClusterWithName(self.projectState),
            self.sessionState.token,
        )
        if "_error" in frameData:
            self.sessionState.loginStatus = False
            self.setEnablementOfAllImportUIElements(False)
            self.setEnablementOfTenantUIElements(False)
            self.setEnablementOfPrintingElements(False)

            return renderWarning(
                "Session timeout - Please login again", self.qgisInterface
            )

        infoMessageAggregator = MessageAggregator()
        result = createPrintComposer(
            self.dlg.printingType_comboBox.currentText(),
            self.dlg.printingFormat_comboBox.currentText(),
            self.dlg.printingScale_comboBox.currentText(),
            self.currentQgisProject,
            self.qgisInterface,
            self.appConfig,
            self.projectState,
            frameData,
            infoMessageAggregator,
        )
        if infoMessageAggregator.hasMessages():
            renderInfo(
                "\n".join(infoMessageAggregator.getMessages()).strip(),
                self.qgisInterface,
            )
        if result.isFailure():
            renderWarning(result.message, self.qgisInterface)

    def handleSelectStyle(self):
        options = QFileDialog.DontResolveSymlinks | QFileDialog.ShowDirsOnly
        startingDirectory = self.appConfig.pluginStyleFolderPath

        selectedStyleFolderPath = QFileDialog.getExistingDirectory(
            None, "Select Style", startingDirectory, options=options
        )

        # otherwise user cancelled
        if selectedStyleFolderPath:
            if not containsFileWithExtension(selectedStyleFolderPath, "qml"):
                renderWarning(
                    "Directory can not be selected as it contains no .qml files",
                    self.qgisInterface,
                )
                return

            selectedStyling = os.path.basename(selectedStyleFolderPath)

            localProjectState = self.projectState.copy()
            localProjectState.currentStyling = selectedStyling

            relativeDir = os.path.relpath(selectedStyleFolderPath, startingDirectory)
            if relativeDir.startswith("..") or "/" in relativeDir:
                # is "local" folder
                localProjectState.overrulingStyleFolderPath = os.path.normpath(
                    self.removeLastDir(selectedStyleFolderPath)
                )
            else:
                localProjectState.overrulingStyleFolderPath = ""

            projectName = self.projectState.getProjectName()
            qgisProjectExistsResult = getExistingProjectFilePath(
                self.appConfig.getProjectDirPath(projectName), projectName
            )
            if qgisProjectExistsResult.isFailure():
                renderWarning(qgisProjectExistsResult.message, self.qgisInterface)
                return

            infoMessageAggregator = MessageAggregator()
            saveResult = saveCurrentProject(
                self.currentQgisProject,
                qgisProjectExistsResult.value,
                localProjectState,
                self.appConfig,
                infoMessageAggregator,
            )

            if infoMessageAggregator.hasMessages():
                renderInfo(
                    "\n".join(infoMessageAggregator.getMessages()).strip(),
                    self.qgisInterface,
                )
            if saveResult.isFailure():
                renderWarning(
                    f"Selecting style failed due to: {saveResult.message}",
                    self.qgisInterface,
                )
                return
            else:
                self.projectState = localProjectState

            changeLayerStyles(
                selectedStyling,
                self.appConfig.getStyleFolderPath(self.projectState),
                self.qgisInterface,
                self.currentQgisProject.mapLayers().values(),
            )

            self.dlg.currentStyle_lineEdit.setText(selectedStyling)

    def removeLastDir(self, path: str):
        return os.path.dirname(path)

    def handleExportQMLs(self):
        options = QFileDialog.DontResolveSymlinks | QFileDialog.ShowDirsOnly
        startingDirectory = self.appConfig.getProjectDirPath(
            self.projectState.getProjectName()
        )

        selectedExportPath = QFileDialog.getExistingDirectory(
            None, "Select directory for export", startingDirectory, options=options
        )

        # otherwise user cancelled
        if selectedExportPath:
            layers = self.currentQgisProject.mapLayers().values()
            vectorLayers = filter(
                lambda layer: layer.type() == QgsMapLayerType.VectorLayer, layers
            )

            for vectorLayer in vectorLayers:
                qmlFilePath = os.path.join(
                    selectedExportPath, f"{vectorLayer.name()}.qml"
                )
                vectorLayer.saveNamedStyle(qmlFilePath)

        renderInfo("Finished exporting qml files", self.qgisInterface)

    def fillLayerSettingComboBox(self):
        comboBox = self.dlg.layer_comboBox
        comboBox.clear()
        layerSettingNames = [
            layerSetting.name for layerSetting in self.appConfig.getRimoLayerSettings()
        ]

        comboBox.addItems(layerSettingNames)

    def setEnablementOfAllImportUIElements(self, active):
        elements = [
            self.dlg.region_comboBox,
            self.dlg.cluster_comboBox,
            self.dlg.siteCluster_comboBox,
            self.dlg.load_pushButton,
            self.dlg.layer_comboBox,
        ]

        self.setEnablement(elements, active)

    def setEnablementOfTenantUIElements(self, active):
        elements = [
            self.dlg.tenant_comboBox,
            self.dlg.selectTenant_pushButton,
        ]

        self.setEnablement(elements, active)

    def setEnablementOfPrintingElements(self, active):
        elements = [
            self.dlg.createPrintComposer_pushButton,
            self.dlg.printingType_comboBox,
            self.dlg.printingFormat_comboBox,
            self.dlg.printingScale_comboBox,
            self.dlg.selectStyle_pushButton,
            self.dlg.exportQMLs_pushButton,
        ]

        self.setEnablement(elements, active)

    def setEnablement(self, elements, enabled):
        for e in elements:
            e.setEnabled(enabled)

    def regionComboBoxChange(self, environment):
        self.dlg.region_comboBox.setEnabled(False)
        self.dlg.cluster_comboBox.clear()
        self.dlg.cluster_comboBox.setEnabled(False)
        self.dlg.siteCluster_comboBox.clear()
        self.dlg.siteCluster_comboBox.setEnabled(False)

        clustersResult = self.loadClusters(environment)
        if clustersResult.isSuccess() and clustersResult.value is not None:
            self.gisDataState.setClusterData(clustersResult.value)
            self.dlg.cluster_comboBox.addItems(self.gisDataState.getClusterNames())

            siteClustersResult = self.loadSiteClusters(environment)
            if siteClustersResult.isSuccess():
                self.gisDataState.setSiteClusterData(siteClustersResult.value)
                siteClusterNames = self.gisDataState.getSiteClusterNames()
                siteClusterNames.sort(key=str.lower)
                siteClusterNames.insert(0, SITE_CLUSTER_DEFAULT)
                self.dlg.siteCluster_comboBox.addItems(siteClusterNames)
            else:
                self.dlg.siteCluster_comboBox.addItems([SITE_CLUSTER_DEFAULT])

        self.dlg.region_comboBox.setEnabled(True)
        self.dlg.cluster_comboBox.setEnabled(True)
        self.dlg.siteCluster_comboBox.setEnabled(True)

    def clusterComboBoxChange(self, environment):
        self.dlg.cluster_comboBox.setEnabled(False)
        self.dlg.siteCluster_comboBox.clear()
        self.dlg.siteCluster_comboBox.setEnabled(False)

        siteClustersResult = self.loadSiteClusters(environment)
        if siteClustersResult.isSuccess():
            self.gisDataState.setSiteClusterData(siteClustersResult.value)
            siteClusterNames = self.gisDataState.getSiteClusterNames()
            siteClusterNames.sort(key=str.lower)
            siteClusterNames.insert(0, SITE_CLUSTER_DEFAULT)
            self.dlg.siteCluster_comboBox.addItems(siteClusterNames)
        else:
            self.dlg.siteCluster_comboBox.addItems([SITE_CLUSTER_DEFAULT])

        self.dlg.cluster_comboBox.setEnabled(True)
        self.dlg.siteCluster_comboBox.setEnabled(True)

    def siteClusterComboBoxChange(self):
        return

    ######################
    # Data loading Methods
    ######################

    def loadRegions(self, environment) -> Result[List[Dict] or None]:
        try:
            regions = getRegions(
                self.appConfig.getRimoUrl(environment),
                self.sessionState.tenantOop,
                self.appConfig.pluginVersion,
                self.sessionState.userOop,
                self.sessionState.token,
            )

            return Result.ok(regions)

        except Exception as e:
            logging.debug(e)
            return Result.fail("Loading region data failed")

    def loadClusters(self, environment) -> Result[List[Dict] or None]:
        try:
            clusters = getClusterData(
                self.appConfig.getRimoUrl(environment),
                self.gisDataState.getOopForRegionWithName(
                    self.dlg.region_comboBox.currentText()
                ),
                self.appConfig.pluginVersion,
                self.sessionState.userOop,
                self.sessionState.token,
            )

            return Result.ok(clusters)

        except Exception as e:
            logging.debug(e)
            return Result.fail("Loading cluster data failed")

    def loadSiteClusters(self, environment) -> Result[List[Dict]]:
        try:
            return Result.ok(
                getSiteClusterData(
                    self.appConfig.getRimoUrl(environment),
                    self.gisDataState.getClusterData()[
                        self.dlg.cluster_comboBox.currentIndex()
                    ]["asOop"],
                    self.appConfig.pluginVersion,
                    self.sessionState.userOop,
                    self.sessionState.token,
                )
            )
        except Exception as e:
            logging.debug(e)
            return Result.fail("Loading site cluster data failed")

    def reloadDataFromRimo(self):
        """Reload Data: Refresh all Rimo Layer after Token is expired"""
        layers = [layer for layer in self.currentQgisProject.mapLayers().values()]
        for layer in layers:
            keys = layer.keywordList()
            if len(keys) > 0 and isNonRevisionRimoLayer(layer):
                url = layer.dataProvider().dataSourceUri()
                r = requests.get(url)
                data = json.loads(r.text)
                if "_error" in data:
                    self.sessionState.loginStatus = False
                    self.setEnablementOfAllImportUIElements(False)
                    self.setEnablementOfTenantUIElements(False)
                    self.setEnablementOfPrintingElements(False)
                    self.dlg.reload_pushButton.setEnabled(False)
                    self.dlg.login_pushButton.setEnabled(True)
                    renderWarning(
                        "Session Time Out - Please Login again", self.qgisInterface
                    )
                    return

                else:
                    layer.dataProvider().reloadData()
                    layer.triggerRepaint()

        renderInfo("Layer reloaded!", self.qgisInterface)
        refreshLayers(
            self.currentQgisProject,
            self.appConfig,
            self.sessionState.token,
            self.qgisInterface,
            self.projectState,
        )

        currentQgisProjectFilePath = createProjectFilePath(
            self.appConfig.getProjectDirPath(self.projectState.getProjectName()),
            self.projectState.getProjectName(),
            self.currentQgisProject.isZipped(),
        )
        infoMessageAggregator = MessageAggregator()
        saveResult = saveCurrentProject(
            self.currentQgisProject,
            currentQgisProjectFilePath,
            self.projectState,
            self.appConfig,
            infoMessageAggregator,
        )
        if infoMessageAggregator.hasMessages():
            renderInfo(
                "\n".join(infoMessageAggregator.getMessages()).strip(),
                self.qgisInterface,
            )
        if saveResult.isFailure():
            renderWarning(saveResult.message, self.qgisInterface)

    def resetPlugin(self):
        self.initState()
        self.initUI()

    def unload(self):
        """Removes the plugin menu item and icon from QGIS GUI."""
        for action in self.actions:
            self.qgisInterface.removePluginMenu(self.tr("&RimoPrinting"), action)
            self.qgisInterface.removeToolBarIcon(action)

    # noinspection PyMethodMayBeStatic
    def tr(self, message):
        """Get the translation for a string using Qt translation API.

        We implement this ourselves since we do not inherit QObject.

        :param message: String for translation.
        :type message: str, QString

        :returns: Translated version of message.
        :rtype: QString
        """
        # noinspection PyTypeChecker,PyArgumentList,PyCallByClass
        return QCoreApplication.translate("RimoPrinting", message)

    def add_action(
        self,
        icon_path,
        text,
        callback,
        enabled_flag=True,
        add_to_menu=True,
        add_to_toolbar=True,
        status_tip=None,
        whats_this=None,
        parent=None,
    ):
        """Add a toolbar icon to the toolbar."""

        icon = QIcon(icon_path)
        action = QAction(icon, text, parent)
        action.triggered.connect(callback)
        action.setEnabled(enabled_flag)
        if status_tip is not None:
            action.setStatusTip(status_tip)

        if whats_this is not None:
            action.setWhatsThis(whats_this)

        if add_to_toolbar:
            self.qgisInterface.addToolBarIcon(action)

        if add_to_menu:
            self.qgisInterface.addPluginToVectorMenu(self.menu, action)

        self.actions.append(action)

        return action

    def restartPlugin(self):
        reloadPlugin(os.path.basename(self.appConfig.pluginDirPath))
        self.dlg.close()
